module Day7.Interpreter (
  buildTree,
  mkdir,
  Directory (..),
  calculateSize,
  filterDirectories,
  sizeOfDirectories,
  sumUp,
) where

import Day7.Parser
import Data.Map (Map)
import qualified Data.Map as Map

data Directory = Directory
  { sub :: Map String Directory
  , files :: Map String Int
  , isRoot :: Bool
} deriving (Eq)


instance Show Directory where
  show d = "DIR " ++ show (Map.toList $ files d) ++ " - " ++ show (Map.toList $ sub d) ++ "\n"

mkdir :: Directory
mkdir = Directory { sub = Map.empty, files = Map.empty, isRoot = False }

rootDirectory :: Directory
rootDirectory = mkdir { isRoot = True  }

buildTree :: [TerminalCommand] -> Directory
buildTree commands = fst (buildTree' rootDirectory commands)
  where
  -- executes terminal commands and returns all remaining commands
  buildTree' :: Directory -> [TerminalCommand] -> (Directory, [TerminalCommand])
  buildTree'  dir [] = (dir, [])
  buildTree' dir (command:cs) =  case command of
    Listing entries ->
      let asFiles = Map.fromList $ toFiles entries
          newDir = dir { files = asFiles }
      in buildTree' newDir cs -- update current directory and return all remaining commands
    In dirName ->
     let subFolders = sub dir
         selectedDir = Map.findWithDefault mkdir dirName subFolders -- select or create folder
         (replacement, rest) = buildTree' selectedDir cs -- recurse down
         updatedSub = Map.insert dirName replacement subFolders -- update folder the map of the current directory
     in buildTree' dir { sub = updatedSub} rest -- continue with updated folder and rest of commands
    Out -> (dir, cs) -- return to previous caller and return all remaining commands
    Root -> if isRoot dir
              then buildTree' dir cs -- return to previous caller as long is current directory is not root
              else (dir, command:cs) -- keep the Root command because it is not yet completed

-- convert to file-pairs and ignore directories
toFiles :: [ListingEntry] -> [(String, Int)]
toFiles ((FileListing name size):rest) = (name, size): toFiles rest
toFiles (_:rest) = toFiles rest
toFiles [] = []

calculateSize :: Directory -> Int
calculateSize dir = sum (Map.elems (files dir)) + sum sizes
  where
    subFolders = Map.elems (sub dir)
    sizes = map calculateSize subFolders

flatten :: Directory -> [(String, Directory)]
flatten dir = flatten' ("/", dir)
  where
  flatten' (name, d) = (name, d) : concatMap flatten' (Map.toList (sub d))

sizeOfDirectories :: Directory -> [(String, Int)]
sizeOfDirectories dir = map withSize allDirectories
  where
    allDirectories = flatten dir
    withSize (name, directory) = (name, calculateSize directory)

filterDirectories :: (Int -> Bool) -> Directory -> [(String, Int)]
filterDirectories predicate dir = filter (\(_, size) -> predicate size) $ sizeOfDirectories dir

sumUp :: [(String, Int)] -> Int
sumUp = sum . map snd
